"""This module provides utilities for working with Google Colab."""
import colaburl
from deprecated import deprecated

_REASON = 'Use the colaburl package instead. pyterrier_alpha.colab will be removed in a future version.'
_VERSION = '0.9.0'

code_url = deprecated(version=_VERSION, reason=_REASON)(colaburl.code_url)
code_html_link = deprecated(version=_VERSION, reason=_REASON)(colaburl.code_html_link)
code_html_form = deprecated(version=_VERSION, reason=_REASON)(colaburl.code_html_form)
code_html = deprecated(version=_VERSION, reason=_REASON)(colaburl.code_html)



"""Fusion methods for combining ranking results."""
from collections import defaultdict
from typing import Optional

import pandas as pd
import pyterrier as pt

import pyterrier_alpha as pta


class RRFusion(pt.Transformer):
  """Reciprocal Rank Fusion between the results from multiple transformers.

  This transformer merges multiple ranking results by computing the reciprocal rank of each document in each ranking,
  and summing them up. The reciprocal rank is computed as 1/(rank + k), where k is a constant. The resulting score is
  used to rank the documents.

  Consider using the :func:`rr_fusion` function if you want to apply fusion outside of a pipeline.

  .. cite.dblp:: conf/sigir/CormackCB09
  """
  def __init__(self,
    *transformers: pt.Transformer,
    k: int = 60,
    num_results: Optional[int] = 1000
  ):
    """Initializes the transformer.

    Args:
      transformers: The transformers to merge.
      k: The constant used in the reciprocal rank computation.
      num_results: The number of results to keep for each query. If None, all results are kept.
    """
    assert len(transformers) > 0
    self.transformers = transformers
    self.k = k
    self.num_results = num_results

  def transform(self, inp: pd.DataFrame) -> pd.DataFrame:
    """Performs the reciprocal rank fusion on the input data."""
    return rr_fusion(*[t(inp) for t in self.transformers], k=self.k, num_results=self.num_results)


def rr_fusion(
  *results: pd.DataFrame,
  k: int = 60,
  num_results: Optional[int] = 1000
) -> pd.DataFrame:
  """Reciprocal Rank Fusion between two ranking result lists.

  Args:
    results: Multiple result frames to merge. At least one frame is required.
    k: The constant used in the reciprocal rank computation.
    num_results: The number of results to keep for each query. If None, all results are kept.

  Consider using :class:`RRFusion` if you want to use this direclty in a pipeline.

  .. cite.dblp:: conf/sigir/CormackCB09
  """
  assert len(results) > 0
  all_qids = defaultdict(list)
  for r in results:
    pta.validate.result_frame(r, extra_columns=['score'])
    pt.model.add_ranks(r)
    r = r[['qid', 'query', 'docno']].assign(score=1/(r['rank'] + k))
    for qid, v in r.groupby('qid'):
      all_qids[qid].append(v)
  res = []
  for qid, qid_frames in all_qids.items():
    merged_frame = qid_frames[0]
    for next_frame in qid_frames[1:]:
      merged_frame = merged_frame.merge(next_frame, how='outer', on=['qid', 'query', 'docno'])
      merged_frame = merged_frame.assign(score=merged_frame['score_x'].fillna(0.) + merged_frame['score_y'].fillna(0.))
      merged_frame.drop(columns=['score_x', 'score_y'], inplace=True)
    pt.model.add_ranks(merged_frame)
    merged_frame.sort_values('rank', ascending=True, inplace=True)
    if num_results is not None: # apply cutoff if present
      res.append(merged_frame[merged_frame['rank'] < num_results])
  return pd.concat(res, ignore_index=True)



"""Module providing the Rank Biased Overlap (RBO) measure."""

from typing import Callable, Iterable, Optional, Tuple

import ir_measures
import pandas as pd


def RBO(other: pd.DataFrame, p: float = 0.99, *, name: Optional[str] = None) -> ir_measures.Measure: # noqa: N802
    """Create an RBO measure from a dataframe of rankings.

    .. versionadded:: 0.3.0

    .. versionchanged:: 0.3.1
        Fixed bug where ``p`` wasn't honored.
    """
    return ir_measures.define(_rbo_wrapper(other, p=p), name=name or f'RBO(p={p})')


def _rbo_wrapper(a: pd.DataFrame, p: float = 0.99) -> Callable:
    # adapted from https://github.com/terrierteam/ir_measures/blob/main/ir_measures/providers/compat_provider.py
    a_q_col = 'query_id' if 'query_id' in a.columns else 'qid'
    a_d_col = 'doc_id' if 'doc_id' in a.columns else 'docno'
    a = a.sort_values(by=[a_q_col, 'score'], ascending=False)
    a = dict(iter(a.groupby(a_q_col)))
    def inner(qrels: pd.DataFrame, b: pd.DataFrame) -> Iterable[Tuple[str, float]]:
        # qrels ignored
        b_q_col = 'query_id' if 'query_id' in b.columns else 'qid'
        b_d_col = 'doc_id' if 'doc_id' in b.columns else 'docno'
        res = {}
        b = b.sort_values(by=[b_q_col, 'score'], ascending=False)
        b = dict(iter(b.groupby(b_q_col)))
        for qid in set(a.keys()) | set(b.keys()):
            ranking = list(a[qid][a_d_col]) if qid in a else []
            ideal = list(b[qid][b_d_col]) if qid in b else []
            ranking_set = set()
            ideal_set = set()
            score = 0.0
            normalizer = 0.0
            weight = 1.0
            for i in range(1000):
                if i < len(ranking):
                    ranking_set.add(ranking[i])
                if i < len(ideal):
                    ideal_set.add(ideal[i])
                score += weight*len(ideal_set.intersection(ranking_set))/(i + 1)
                normalizer += weight
                weight *= p
            res[qid] = score/normalizer
        return res.items()
    return inner


def rbo(a: pd.DataFrame, b: pd.DataFrame, p: float = 0.99) -> Iterable[Tuple[str, float]]:
    """Calculate the Rank Biased Overlap between two rankings.

    .. versionadded:: 0.3.0
    .. versionchanged:: 0.12.5
        Fixed bug where b wasn't passed properly
    """
    return _rbo_wrapper(a, p)(None, b)



"""PyTerrier artifact module."""
import json
from enum import Enum
from pathlib import Path
from typing import Any, Optional, Union

from pyterrier._artifact import Artifact, _hf_url_resolver

import pyterrier_alpha as pta

__all__ = [
    'Artifact',
    'ArtifactBuilder',
    'ArtifactBuilderMode',
    'from_url',
    'from_dataset',
    '_hf_url_resolver',
]

from_url = Artifact.from_url
from_dataset = Artifact.from_dataset


class ArtifactBuilderMode(Enum):
    """Enumeration of artifact builder modes."""
    create = 'create'
    overwrite = 'overwrite'
    append = 'append'


class ArtifactBuilder:
    """Context manager for building artifacts."""
    def __init__(
        self,
        artifact: Optional[Artifact] = None,
        *,
        mode: Union[ArtifactBuilderMode, str] = 'create',
        path: Optional[Union[str, Path]] = None,
        type: Optional[str] = None,
        format: Optional[str] = None,
        package_hint: Optional[str] = None,
    ) -> None:
        """Context manager for building artifacts."""
        if artifact is None:
            assert path is not None, "path must be provided if artifact isn't"
            assert type is not None, "type must be provided if artifact isn't"
            assert format is not None, "format must be provided if artifact isn't"
            assert package_hint is not None, "package_hint must be provided if artifact isn't"

        self.mode = ArtifactBuilderMode(mode)
        self.path = Path(path or artifact.path)

        if type is None or format is None:
            type, format = pta.inspect.artifact_type_format(artifact)

        if package_hint is None:
            if hasattr(artifact, 'ARTIFACT_PACKAGE_HINT'):
                package_hint = artifact.ARTIFACT_PACKAGE_HINT
            else:
                package_hint = artifact.__class__.__module__.split('.')[0]

        self.metadata = {
            'type': type,
            'format': format,
            'package_hint': package_hint,
        }

    def __enter__(self) -> 'ArtifactBuilder':
        # TODO: check if the path exists and either:
        if self.mode == ArtifactBuilderMode.create:
            if self.path.exists():
                raise FileExistsError(f'{str(self.path)} already exists.')
            self.path.mkdir(parents=True)
        elif self.mode == ArtifactBuilderMode.overwrite:
            if self.path.exists():
                raise NotImplementedError()
                # This might be a bit tricky... Do we recursively delete the entire
                # directory if it exists? What if the user accidently enters a path
                # such as '/'? Do we need some sort of check or warning if the path
                # is very large? Hmmm...
            else:
                self.path.mkdir(parents=True)
        elif self.mode == ArtifactBuilderMode.append:
            raise NotImplementedError()

        return self

    def __exit__(self, exc_type: Any, exc_value: Any, traceback: Any) -> None:
        if exc_type is not None:
            # TODO clean up
            pass
        else:
            # Log the artifact metadata
            meta_path = self.path / 'pt_meta.json'
            try:
                with open(meta_path, 'wt') as fout:
                    json.dump(self.metadata, fout)
            except:
                if meta_path.exists():
                    meta_path.unlink()
                raise



"""Extension I/O utilities for PyTerrier."""

import io
import os
import shutil
import tempfile
import urllib
from abc import ABC, abstractmethod
from contextlib import ExitStack, contextmanager
from hashlib import sha256
from importlib.metadata import EntryPoint
from importlib.metadata import entry_points as eps
from typing import IO, BinaryIO, Callable, Iterable, Optional, Tuple

import pyterrier as pt
from deprecated import deprecated

DEFAULT_CHUNK_SIZE = 16_384 # 16kb


@contextmanager
def _finalized_open_base(path: str, mode: str, open_fn: Callable) -> io.IOBase:
    assert mode in ('b', 't') # must supply either binary or text mode
    prefix = f'.{os.path.basename(path)}.tmp.'
    dirname = os.path.dirname(path)
    path_tmp = None
    try:
        fd, path_tmp = tempfile.mkstemp(prefix=prefix, dir=dirname)
        os.close(fd) # mkstemp returns a low-level file descriptor... Close it and re-open the file the normal way
        with open_fn(path_tmp, f'w{mode}') as fout:
            yield fout
        os.chmod(path_tmp, 0o666) # default file umask
    except:
        if path_tmp is not None:
            os.remove(path_tmp)
        raise

    os.replace(path_tmp, path)


def finalized_open(path: str, mode: str) -> IO:
    """Opens a file for writing, but reverts it if there was an error in the process.

    Args:
        path(str): Path of file to open
        mode(str): Either t or b, for text or binary mode

    Example:
        Returns a contextmanager that provides a file object, so should be used in a "with" statement. E.g.::

            with pta.io.finalized_open("file.txt", "t") as f:
                f.write("some text")
            # file.txt exists with contents "some text"

        If there is an error when writing, the file is reverted::

            with pta.io.finalized_open("file.txt", "t") as f:
                f.write("some other text")
                raise Exception("an error")
            # file.txt remains unchanged (if existed, contents unchanged; if didn't exist, still doesn't)
    """
    return _finalized_open_base(path, mode, open)


@contextmanager
def finalized_directory(path: str) -> str:
    """Creates a directory, but reverts it if there was an error in the process."""
    prefix = f'.{os.path.basename(path)}.tmp.'
    dirname = os.path.dirname(path)
    try:
        path_tmp = tempfile.mkdtemp(prefix=prefix, dir=dirname)
        yield path_tmp
        os.chmod(path_tmp, 0o777) # default directory umask
    except:
        try:
            shutil.rmtree(path_tmp)
        except:
            raise
        raise

    os.replace(path_tmp, path)


def download(url: str, path: str, *, expected_sha256: str = None, verbose: bool = True) -> None:
    """Downloads a file from a URL to a local path."""
    with finalized_open(path, mode='b') as fout, \
         download_stream(url, expected_sha256=expected_sha256, verbose=verbose) as fin:
        while chunk := fin.read1():
            fout.write(chunk)


@contextmanager
def download_stream(url: str, *, expected_sha256: Optional[str] = None, verbose: bool = True) -> io.IOBase:
    """Downloads a file from a URL to a stream."""
    with ExitStack() as stack:
        fin = stack.enter_context(urllib.request.urlopen(url))
        if fin.status != 200:
            raise OSError(f'Unhandled status code: {fin.status}')

        if verbose:
            total = int(fin.headers.get('Content-Length', 0)) or None
            fin = stack.enter_context(TqdmReader(fin, total=total, desc=url))

        if expected_sha256 is not None:
            fin = stack.enter_context(HashReader(fin, expected=expected_sha256))

        yield fin


@contextmanager
def open_or_download_stream(
    path_or_url: str,
    *,
    expected_sha256: Optional[str] = None,
    verbose: bool = True
) -> io.IOBase:
    """Opens a file or downloads a file from a URL to a stream."""
    if path_or_url.startswith('http://') or path_or_url.startswith('https://'):
        with download_stream(path_or_url, expected_sha256=expected_sha256, verbose=verbose) as fin:
            yield fin
    elif os.path.isfile(path_or_url):
        with ExitStack() as stack:
            fin = stack.enter_context(open(path_or_url, 'rb'))

            if verbose:
                total = os.path.getsize(path_or_url)
                fin = stack.enter_context(TqdmReader(fin, total=total, desc=path_or_url))

            if expected_sha256 is not None:
                fin = stack.enter_context(HashReader(fin, expected=expected_sha256))

            yield fin
    else:
        raise OSError(f'path or url {path_or_url!r} not found')


class _NosyReader(io.BufferedIOBase, ABC):
    def __init__(self, reader: io.IOBase):
        self.reader = reader
        self.seek = self.reader.seek
        self.tell = self.reader.tell
        self.seekable = self.reader.seekable
        self.readable = self.reader.readable
        self.writable = self.reader.writable
        self.flush = self.reader.flush
        self.isatty = self.reader.isatty

    @abstractmethod
    def on_data(self, data: bytes) -> None:
        pass

    def read1(self, size: int = -1) -> bytes:
        if size == -1:
            size = DEFAULT_CHUNK_SIZE
        chunk = self.reader.read1(min(size, DEFAULT_CHUNK_SIZE))
        self.on_data(chunk)
        return chunk

    def read(self, size: int = -1) -> bytes:
        if size == -1:
            size = DEFAULT_CHUNK_SIZE
        chunk = self.reader.read(min(size, DEFAULT_CHUNK_SIZE))
        self.on_data(chunk)
        return chunk

    def close(self) -> None:
        self.reader.close()


class _NosyWriter(io.BufferedIOBase, ABC):
    def __init__(self, writer: io.IOBase):
        self.writer = writer
        self.seek = self.writer.seek
        self.tell = self.writer.tell
        self.seekable = self.writer.seekable
        self.readable = self.writer.readable
        self.writable = self.writer.writable
        self.flush = self.writer.flush
        self.isatty = self.writer.isatty
        self.close = self.writer.close
        self.sha256 = sha256()

    @abstractmethod
    def on_data(self, data: bytes) -> None:
        pass

    def write(self, data: bytes) -> None:
        self.writer.write(data)
        self.on_data(data)

    def replace_writer(self, writer: io.IOBase) -> None:
        self.writer = writer
        self.seek = self.writer.seek
        self.tell = self.writer.tell
        self.seekable = self.writer.seekable
        self.readable = self.writer.readable
        self.writable = self.writer.writable
        self.flush = self.writer.flush
        self.isatty = self.writer.isatty
        self.close = self.writer.close


class HashReader(_NosyReader):
    """A reader that computes the sha256 hash of the data read."""
    def __init__(self, reader: io.IOBase, *, hashfn: Callable = sha256, expected: Optional[str] = None):
        """Create a HashReader."""
        super().__init__(reader)
        self.hash = hashfn()
        self.expected = expected

    def on_data(self, data: bytes) -> None:
        """Called when data is read."""
        self.hash.update(data)

    def hexdigest(self) -> str:
        """Return the hexdigest of the hash."""
        return self.hash.hexdigest()

    def close(self) -> None:
        """Close the reader and check the hash."""
        self.reader.close()
        if self.expected is not None:
            if self.expected.lower() != self.hexdigest():
                raise ValueError(f'Expected sha256 {self.expected!r} but found {self.hexdigest()!r}')


class HashWriter(_NosyWriter):
    """A writer that computes the sha256 hash of the data written."""
    def __init__(self, writer: io.IOBase, *, hashfn: Callable = sha256):
        """Create a HashWriter."""
        super().__init__(writer)
        self.hash = hashfn()

    def on_data(self, data: bytes) -> None:
        """Called when data is written."""
        self.hash.update(data)

    def hexdigest(self) -> str:
        """Return the hexdigest of the hash."""
        return self.hash.hexdigest()


class TqdmReader(_NosyReader):
    """A reader that displays a progress bar."""
    def __init__(self, reader: io.IOBase, *, total: int = None, desc: str = None, disable: bool = False):
        """Create a TqdmReader."""
        super().__init__(reader)
        import pyterrier as pt
        self.pbar = pt.tqdm(total=total, desc=desc, unit="B", unit_scale=True, unit_divisor=1024, disable=disable)

    def on_data(self, data: bytes) -> None:
        """Called when data is read."""
        self.pbar.update(len(data))

    def close(self) -> None:
        """Close the reader and the progress bar."""
        super().close()
        self.reader.close()


class CallbackReader(_NosyReader):
    """A reader that calls a callback with the data read."""
    def __init__(self, reader: io.IOBase, callback: Callable):
        """Create a CallbackReader."""
        super().__init__(reader)
        self.callback = callback

    def on_data(self, data: bytes) -> None:
        """Called when data is read."""
        self.callback(data)


class MultiReader(io.BufferedIOBase):
    """A reader that reads from multiple readers in sequence."""
    def __init__(self, readers: Iterable[BinaryIO]):
        """Create a MultiReader."""
        self.readers = readers
        self._reader = next(self.readers)
        self.reader = self._reader.__enter__()
        self.seek = self.reader.seek
        self.tell = self.reader.tell
        self.seekable = self.reader.seekable
        self.readable = self.reader.readable
        self.writable = self.reader.writable
        self.flush = self.reader.flush
        self.isatty = self.reader.isatty
        self.close = self.reader.close

    def read1(self, size: int = -1) -> bytes:
        """Read a single chunk of data."""
        if size == -1:
            size = DEFAULT_CHUNK_SIZE
        chunk = self.reader.read1(min(size, DEFAULT_CHUNK_SIZE))
        if len(chunk) == 0:
            self.reader.close()
            try:
                self._reader = next(self.readers)
            except StopIteration:
                self._reader = None
                self.reader = None
                return chunk
            self.reader = self._reader.__enter__()
            self.pbar = self.reader.pbar
            self.seek = self.reader.seek
            self.tell = self.reader.tell
            self.seekable = self.reader.seekable
            self.readable = self.reader.readable
            self.writable = self.reader.writable
            self.flush = self.reader.flush
            self.isatty = self.reader.isatty
            self.close = self.reader.close
            chunk = self.reader.read1(min(size, DEFAULT_CHUNK_SIZE))
        return chunk

    def read(self, size: int = -1) -> bytes:
        """Read data."""
        chunk = b''
        if size == -1:
            size = DEFAULT_CHUNK_SIZE
        while len(chunk) < size and self.reader is not None:
            chunk += self.reader.read(size - len(chunk))
            if len(chunk) < size:
                self.reader.close()
                try:
                    self._reader = next(self.readers)
                except StopIteration:
                    self._reader = None
                    self.reader = None
                    return chunk
                self.reader = self._reader.__enter__()
                self.pbar = self.reader.pbar
                self.seek = self.reader.seek
                self.tell = self.reader.tell
                self.seekable = self.reader.seekable
                self.readable = self.reader.readable
                self.writable = self.reader.writable
                self.flush = self.reader.flush
                self.isatty = self.reader.isatty
                self.close = self.reader.close
        return chunk


def path_is_under_base(path: str, base: str) -> bool:
    """Returns True if the path is under the base directory."""
    return os.path.realpath(os.path.abspath(os.path.join(base, path))).startswith(os.path.realpath(base))


def byte_count_to_human_readable(byte_count: float) -> str:
    """Converts a byte count to a human-readable string."""
    units = ['B', 'KB', 'MB', 'GB', 'TB']
    while byte_count > 1024 and len(units) > 1:
        byte_count /= 1024
        units = units[1:]
    if units[0] == 'B':
        return f'{byte_count:.0f} {units[0]}'
    return f'{byte_count:.1f} {units[0]}'


def entry_points(group: str) -> Tuple[EntryPoint, ...]:
    """Returns the entry points for a given group."""
    try:
        return tuple(eps(group=group))
    except TypeError:
        return tuple(eps().get(group, tuple()))


_REASON = ('python-terrier>=0.11.0 added `pyterrier.io.pyterrier_home`. Use this instead. '
           '`pyterrier_alpha.io.pyterrier_home` will be removed in a future version.')
_VERSION = '0.10.0'
try:
    # Moved to pyterrier core in 0.11.0
    pyterrier_home = deprecated(version=_VERSION, reason=_REASON)(pt.io.pyterrier_home)
except AttributeError:
    @deprecated(version=_VERSION, reason=_REASON)
    def pyterrier_home() -> str:
        """Returns the PyTerrier home directory."""
        if "PYTERRIER_HOME" in os.environ:
            home = os.environ["PYTERRIER_HOME"]
        else:
            home = os.path.expanduser('~/.pyterrier')
        if not os.path.exists(home):
            os.makedirs(home)
        return home



"""Validation utilities for checking the input to transformers."""

from types import TracebackType
from typing import List, Optional, Type

import pandas as pd

from pyterrier_alpha.utils import PeekableIter


class _TransformerMode:
    def __init__(self, missing_columns: List[str], extra_columns: List[str], mode_name: Optional[str] = None):
        self.missing_columns = missing_columns
        self.extra_columns = extra_columns
        self.mode_name = mode_name

    def __str__(self):
        return f'{self.mode_name} (missing: {self.missing_columns}, extra: {self.extra_columns})'

    def __repr__(self):
        return f'TransformerMode(missing_columns={self.missing_columns!r}, ' \
               f'extra_columns={self.extra_columns!r}, ' \
               f'mode_name={self.mode_name!r})'


class InputValidationError(KeyError):
    """Exception raised when input validation fails."""
    def __init__(self, message: str, modes: List[_TransformerMode]):
        """Create an InputValidationError."""
        assert len(modes) > 0
        super().__init__(message)
        self.modes = modes

    def __str__(self):
        return self.args[0] + ' ' + str(self.modes)

    def __repr__(self):
        return f'InputValidationError({self.args[0]!r}, {self.modes!r})'


def columns(inp: pd.DataFrame,
            *,
            includes: Optional[List[str]] = None,
            excludes: Optional[List[str]] = None) -> None:
    """Check that the input frame has the expected columns.

    Raises:
        InputValidationError: If is missing any of the `includes` columns, or has any of the `excludes` columns.
    """
    with any(inp) as v:
        v.columns(includes=includes, excludes=excludes)


def query_frame(inp: pd.DataFrame, extra_columns: Optional[List[str]] = None) -> None:
    """Check that the input frame is a valid query frame.

    Raises:
        InputValidationError: If the input frame is not a valid query frame.
    """
    with any(inp) as v:
        v.query_frame(extra_columns)



def result_frame(inp: pd.DataFrame, extra_columns: Optional[List[str]] = None) -> None:
    """Check that the input frame is a valid result frame.

    Raises:
        InputValidationError: If the input frame is not a valid result frame.
    """
    with any(inp) as v:
        v.result_frame(extra_columns)


def document_frame(inp: pd.DataFrame, extra_columns: Optional[List[str]] = None) -> None:
    """Check that the input frame is a valid document frame.

    Raises:
        InputValidationError: If the input frame is not a valid document frame.
    """
    with any(inp) as v:
        v.document_frame(extra_columns)


def columns_iter(inp: pd.DataFrame,
            *,
            includes: Optional[List[str]] = None,
            excludes: Optional[List[str]] = None) -> None:
    """Check that the input frame has the expected columns.

    Raises:
        InputValidationError: If is missing any of the `includes` columns, or has any of the `excludes` columns.
    """
    with any_iter(inp) as v:
        v.columns(includes=includes, excludes=excludes)


def any(inp: pd.DataFrame) -> '_ValidationContextManager':
    """Create a validation context manager for a DataFrame."""
    return _ValidationContextManager(inp)


def any_iter(inp: PeekableIter) -> '_IterValidationContextManager':
    """Create a validation context manager for an iterator."""
    if not isinstance(inp, PeekableIter):
        raise AttributeError('inp is not peekable. Run the following before calling this function.\n'
                             'inp = pta.utils.peekable(inp) # !! IMPORTANT: you must re-assign the input to peekable '
                             '(not just pass it in), otherwise you will skip the first record !!')
    return _IterValidationContextManager(inp)


class _ValidationContextManager:
    """Context manager for validating the input to transformers."""
    def __init__(self, inp: pd.DataFrame):
        """Create a ValidationContextManager for the given DataFrame."""
        self.inp = inp
        self.mode = None
        self.attempts = 0
        self.errors = []

    def __enter__(self):
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc: Optional[BaseException],
        traceback: Optional[TracebackType]
    ) -> Optional[bool]:

        if exc_type is not None:
            return False # the captured exception takes priority

        if self.attempts > 0 and self.attempts == len(self.errors):
            raise InputValidationError("DataFrame does not match required columns for this transformer.", self.errors)

    def columns(self,
                *,
                includes: Optional[List[str]] = None,
                excludes: Optional[List[str]] = None,
                mode: str = None) -> bool:
        """Check that the input frame has the ``includes`` columns and doesn't have the ``excludes`` columns."""
        includes = includes if includes is not None else []
        excludes = excludes if excludes is not None else []
        missing_columns = set(includes) - set(self.inp.columns)
        extra_columns = set(excludes) & set(self.inp.columns)
        self.attempts += 1

        if missing_columns or extra_columns:
            self.errors.append(_TransformerMode(
                missing_columns=[c for c in includes if c in missing_columns],
                extra_columns=[c for c in excludes if c in extra_columns],
                mode_name=mode,
            ))
            return False

        if self.mode is None and mode is not None:
            self.mode = mode

        return True

    def query_frame(self, extra_columns: Optional[List[str]] = None, mode: str = None) -> bool:
        """Check that the input frame is a valid query frame, with optional extra columns."""
        extra_columns = list(extra_columns) if extra_columns is not None else []
        return self.columns(includes=['qid'] + extra_columns, excludes=['docno'], mode=mode)

    def result_frame(self, extra_columns: Optional[List[str]] = None, mode: str = None) -> bool:
        """Check that the input frame is a valid result frame, with optional extra columns."""
        extra_columns = list(extra_columns) if extra_columns is not None else []
        return self.columns(includes=['qid', 'docno'] + extra_columns, mode=mode)

    def document_frame(self, extra_columns: Optional[List[str]] = None, mode: str = None) -> bool:
        """Check that the input frame is a valid document frame, with optional extra columns."""
        extra_columns = list(extra_columns) if extra_columns is not None else []
        return self.columns(includes=['docno'] + extra_columns, excludes=['qid'], mode=mode)


_EMPTY_ITER = object()

class _IterValidationContextManager:
    def __init__(self, inp: PeekableIter):
        try:
            self.sample_cols = set(inp.peek().keys())
        except StopIteration:
            self.sample_cols = _EMPTY_ITER
        self.mode = None
        self.attempts = 0
        self.errors = []

    def __enter__(self):
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc: Optional[BaseException],
        traceback: Optional[TracebackType]
    ) -> Optional[bool]:

        if exc_type is not None:
            return False # the captured exception takes priority

        if self.attempts > 0 and self.attempts == len(self.errors):
            raise InputValidationError("Input does not match required columns for this transformer.", self.errors)

    def columns(self,
                *,
                includes: Optional[List[str]] = None,
                excludes: Optional[List[str]] = None,
                mode: str = None) -> bool:
        self.attempts += 1
        includes = includes if includes is not None else []
        excludes = excludes if excludes is not None else []
        if self.sample_cols == _EMPTY_ITER:
            self.errors.append(_TransformerMode(
                missing_columns=list(includes),
                extra_columns=[],
                mode_name=mode,
            ))
            return False
        missing_columns = set(includes) - self.sample_cols
        extra_columns = set(excludes) & self.sample_cols

        if missing_columns or extra_columns:
            self.errors.append(_TransformerMode(
                missing_columns=[c for c in includes if c in missing_columns],
                extra_columns=[c for c in excludes if c in extra_columns],
                mode_name=mode,
            ))
            return False

        if self.mode is None and mode is not None:
            self.mode = mode

        return True

    def empty(self, *, mode: str = 'empty'):
        self.attempts += 1
        if self.sample_cols != _EMPTY_ITER:
            self.errors.append(_TransformerMode(
                missing_columns=[],
                extra_columns=[],
                mode_name=mode,
            ))
            return False

        if self.mode is None and mode is not None:
            self.mode = mode
        return True



import unittest
import pandas as pd
import pyterrier as pt
import pyterrier_alpha as pta


class TestFusion(unittest.TestCase):

    def test_rrf_transformer(self):
        index = pt.terrier.TerrierIndex.from_hf('pyterrier/vaswani.terrier')
        topics = pt.get_dataset('irds:vaswani').get_topics()
        res_a = index.bm25(k1=1.0, b=1.0)(topics)
        res_b = index.bm25(k1=5.0, b=0.0)(topics)
        res_c = index.bm25(k1=2.5, b=0.5)(topics)
        results = pta.fusion.rr_fusion(res_a, res_b, res_c)

    def test_rrf_transformer(self):
        index = pt.terrier.TerrierIndex.from_hf('pyterrier/vaswani.terrier')
        topics = pt.get_dataset('irds:vaswani').get_topics()
        pipeline_a = index.bm25(k1=1.0, b=1.0)
        pipeline_b = index.bm25(k1=5.0, b=0.0)
        pipeline_c = index.bm25(k1=2.5, b=0.5)
        rrf = pta.RRFusion(pipeline_a, pipeline_b, pipeline_c)
        results = rrf(topics)



"""Module that provides utility functions."""

from typing import Any, Iterable, Iterator, Union

_NO_BUFFER = object()

class PeekableIter:
    """An iterator that allows peeking at the next element."""
    def __init__(self, base: Union[Iterator, Iterable]):
        """Create a PeekableIter from an iterator or iterable."""
        self.base = iter(base)
        self._buffer = _NO_BUFFER

    def __getattr__(self, attr: str):
        return getattr(self.base, attr)

    def __next__(self):
        if self._buffer != _NO_BUFFER:
            n = self._buffer
            self._buffer = _NO_BUFFER
            return n
        return next(self.base)

    def __iter__(self):
        return self

    def peek(self) -> Any:
        """Return the next element without consuming it."""
        if self._buffer == _NO_BUFFER:
            self._buffer = next(self.base)
        return self._buffer


def peekable(it: Union[Iterator, Iterable]) -> PeekableIter:
    """Create a PeekableIter from an iterator or iterable."""
    return PeekableIter(it)



"""Module for inspecting pyterrier objects."""
from typing import List, Optional, Protocol, Tuple, Type, Union, runtime_checkable

import pandas as pd
import pyterrier as pt

import pyterrier_alpha as pta


class InspectError(TypeError):
    """Base exception for inspection errors."""
    pass


def artifact_type_format(
    artifact: Union[Type, 'pta.Artifact'],
    *,
    strict: bool = True,
) -> Optional[Tuple[str, str]]:
    """Returns the type and format of the specified artifact.

    These values are sourced by either the ``ARTIFACT_TYPE`` and ``ARTIFACT_FORMAT`` constants of the artifact, or (if
    these are not available) by matching on the entry points.

    Args:
        artifact: The artifact to inspect.
        strict: If True, raises an error if the type or format could not be determined. If False, returns None
            in these cases.

    Returns:
        A tuple containing the artifact's type and format.

    Raises:
        InspectError: If the artifact's type or format could not be determined and ``strict==True``.
    """
    artifact_type, artifact_format = None, None

    # Source #1: ARTIFACT_TYPE and ARTIFACT_FORMAT constants
    if hasattr(artifact, 'ARTIFACT_TYPE') and hasattr(artifact, 'ARTIFACT_FORMAT'):
        artifact_type = artifact.ARTIFACT_TYPE
        artifact_format = artifact.ARTIFACT_FORMAT

    # Source #2: entry point name
    if artifact_type is None or artifact_format is None:
        for entry_point in pta.io.entry_points('pyterrier.artifact'):
            if artifact.__module__.split('.')[0] != entry_point.value.split(':')[0].split('.')[0]:
                continue # only try loading entry points that share the same top-level module
            entry_point_cls = entry_point.load()
            if isinstance(artifact, type) and artifact == entry_point_cls or isinstance(artifact, entry_point_cls):
                artifact_type, artifact_format = entry_point.name.split('.', 1)
                break

    if artifact_type is None or artifact_format is None:
        if strict:
            raise InspectError(f'{artifact} does not provide type and format (either as constants or via entry point)')
        else:
            return None

    return artifact_type, artifact_format


@runtime_checkable
class ProvidesTransformerOutputs(Protocol):
    """Protocol for transformers that provide a ``transform_outputs`` method."""
    def transform_outputs(self, input_columns: List[str]) -> List[str]:
        """Returns a list of the output columns present given the ``input_columns``."""


def transformer_outputs(
    transformer: pt.Transformer,
    input_columns: List[str],
    *,
    strict: bool = True,
) -> Optional[List[str]]:
    """Infers the output columns for a transformer based on the inputs.

    The method first checks if the transformer provides a ``transform_outputs`` method. If it does, this method is
    called and the result is returned. If the transformer does not provide this method, the method tries to infer the
    outputs by calling the transformer with an empty DataFrame.

    Args:
        transformer: An instance of the transformer to inspect.
        input_columns: A list of the columns present in the input frame.
        strict: If True, raises an error if the transformer cannot be inferred or are not accepted. If False, returns
            None in these cases.

    Returns:
        A list of the columns present in the output for ``transformer`` given ``input_columns``.

    Raises:
        InspectError: If the artifact's type or format could not be determined and ``strict==True``.

    .. versionadded:: 0.11.0
    """
    if isinstance(transformer, ProvidesTransformerOutputs):
        try:
            return transformer.transform_outputs(input_columns)
        except Exception as ex:
            if strict:
                raise InspectError(f"Cannot determine outputs for {transformer} with inputs: {input_columns}") from ex
            else:
                return None

    try:
        res = transformer.transform(pd.DataFrame(columns=input_columns))
        return list(res.columns)
    except Exception as ex:
        if strict:
            raise InspectError(f"Cannot determine outputs for {transformer} with inputs: {input_columns}") from ex
        else:
            return None



"""Module providing a function that calculates a string representation function for transformers."""

import inspect
from typing import Any


def transformer_repr(self: Any) -> str:
    """Return a string representation of a transformer instance.

    .. versionadded:: 0.3.0

    .. versionchanged:: 0.4.2
        Prioritize fields with ``_name`` above ``name``

    .. versionchanged:: 0.12.1
        Ignore verbose
    """
    cls = self.__class__
    init = self.__init__
    signature = inspect.signature(init)
    mode = 'pos'
    args = []
    for p in signature.parameters.values():
        if p.kind not in (inspect.Parameter.POSITIONAL_ONLY, inspect.Parameter.POSITIONAL_OR_KEYWORD):
            mode = 'kwd'
        try:
            val = getattr(self, f'_{p.name}')
        except AttributeError:
            val = getattr(self, p.name)
        if val != p.default and p.name != 'verbose':
            args.append(f'{p.name}={val!r}' if mode == 'kwd' else repr(val))
        else:
            mode = 'kwd' # skip a parameter, force keyword mode
    return cls.__name__ + '(' + ', '.join(args) + ')'
